/* genetic.c -- genetic algorithm demo; no RTcmix interface */
#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/DrawingA.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <unistd.h>

#define  ABS(x) ((x < 0) ? (-x) : (x))

Widget topLevel, bboard;
Widget theDraw;

XtAppContext app;

GC gc;
Colormap cmap;
XColor bground,zerox,sinetarg,line,unused;

#define NPROGENY 100
#define LINEWIDTH 7
#define ARRSIZE 100
#define NBEST 10
#define AMP 300
#define PI2 6.2831853
#define VARAMT 20

int drawwidth,drawheight;

double sinearr[ARRSIZE];
double kids[NPROGENY][ARRSIZE];
double survivors[NBEST][ARRSIZE];
long bestones[NBEST][2]; /* 0 == index; 1 == error */

int
main(argc,argv)
int argc;
char *argv[];
{
	float rrand();
	void srrand();
	void GoDraw();
	double sin();
	int i,j;

	srrand(78);

/* and now the graphics stuff */
	topLevel=XtVaAppInitialize(&app,"genetic",
		NULL,0,&argc,argv,NULL,NULL);

	/* set up the main window */
	drawwidth = LINEWIDTH * ARRSIZE;
	drawheight = AMP * 2;
	bboard = XtVaCreateManagedWidget("bboard",
		xmBulletinBoardWidgetClass, topLevel,
		XmNwidth, drawwidth, XmNheight, drawheight,
		XmNmarginWidth,0,XmNmarginHeight,0,NULL);

	/* set up the colors */
	cmap = DefaultColormapOfScreen(XtScreen(bboard));
	XAllocNamedColor(XtDisplay(bboard),cmap,"powderblue",&bground,&unused);
	XAllocNamedColor(XtDisplay(bboard),cmap,"seagreen",&zerox,&unused);
	XAllocNamedColor(XtDisplay(bboard),cmap,"rosybrown",&sinetarg,&unused);
	XAllocNamedColor(XtDisplay(bboard),cmap,"yellow2",&line,&unused);
	gc = XCreateGC(XtDisplay(bboard),
		RootWindowOfScreen(XtScreen(bboard)), 0, NULL);

	/* set up the drawing area in the main window */
	theDraw = XtVaCreateManagedWidget("theDraw",
		xmDrawingAreaWidgetClass, bboard,
		XmNwidth, drawwidth, XmNheight, drawheight,
		XmNbackground, bground.pixel, NULL);
	XtAddCallback(theDraw, XmNinputCallback, GoDraw, NULL);

	/* set up the sine wave "fitness" function */
	for (i = 0; i < ARRSIZE; i++) {
		sinearr[i] = sin((double)(PI2 * ((float)i/(float)ARRSIZE)));
		sinearr[i] *= (double)AMP;
	}

	/* calculate the first set of NBEST "parents" */
	for (i = 0; i < NBEST; i++) {
		for (j = 0; j < ARRSIZE; j++) {
			survivors[i][j] = (double)rrand() * (double)AMP;
		}
	}

	XtManageChild(bboard);
	XtManageChild(theDraw);
	XtRealizeWidget(topLevel);

	XtAppMainLoop(app);
}



void GoDraw(w, client, call)
Widget w;
XtPointer client, call;
{
	float rrand();
	double err,terr;
	int i,j,k,ifact,jfact;
	int yval;
	int genno;


genno = 0;
while (1) {
usleep(100000);
printf("generation #: %d\n",genno++);

/* generate NPROGENY kids from each of the NBEST */
	jfact = NPROGENY/NBEST;
	for (i = 0; i < NBEST; i++) {
		ifact = i*jfact;
		for (j = 0; j < jfact; j++) {
			for (k = 0; k < ARRSIZE; k++) {
				kids[ifact+j][k] = survivors[i][k] + ((double)rrand() * (double)VARAMT);
				if (kids[ifact+j][k] > (double)AMP) kids[ifact+j][k] = AMP;
				if (kids[ifact+j][k] < (double)(0.0-AMP)) kids[ifact+j][k] = -AMP;
			}
		}
	}

/* now we gotta find which ones are the best */
	/* zero out the error array */
	for (i = 0; i < NBEST; i++) {
		bestones[i][0] = 0;
		bestones[i][1] = 999999999;
	}

	/* and now search for the best */
	for (i = 0; i < NPROGENY; i++) {
		/* calculate the error for this one */
		err = 0.0;
		for (j = 0; j < ARRSIZE; j++) {
			terr = ABS((sinearr[j]-kids[i][j]));
			err += terr;
		}
		/* now let's see if it's one of the best */
		for (j = 0; j < NBEST; j++) {
			if (err < (double)bestones[j][1]) {
				k = NBEST;
				/* shuffle the others down */
				while(--k > j) {
					bestones[k][0] = bestones[k-1][0];
					bestones[k][1] = bestones[k-1][1];
				}
				bestones[j][0] = (long)i;
				bestones[j][1] = (long)err;
				j = NBEST;
			}
		}
	}


/* now we plot out the best ones and transfer them to the survivors */
	for (i = 0; i < NBEST; i++) {
		for (j = 0; j < ARRSIZE; j++) {
			XSetForeground(XtDisplay(theDraw), gc, bground.pixel);
			XFillRectangle(XtDisplay(theDraw),XtWindow(theDraw),gc,
				j*LINEWIDTH,0,LINEWIDTH,drawheight);

			survivors[i][j] = kids[(int)bestones[i][0]][j];

			/* and let's plot this puppy! */
			XSetForeground(XtDisplay(theDraw), gc, line.pixel);
			yval = survivors[i][j];
			yval = AMP - yval;
			if (yval < AMP)
				XFillRectangle(XtDisplay(theDraw),XtWindow(theDraw),gc,j*LINEWIDTH,yval,LINEWIDTH,AMP-yval);
			if (yval > AMP)
				XFillRectangle(XtDisplay(theDraw),XtWindow(theDraw),gc,j*LINEWIDTH,AMP,LINEWIDTH,yval-AMP);

			/* and the sine target */
			XSetForeground(XtDisplay(theDraw), gc, sinetarg.pixel);
			yval = sinearr[j];
			yval = AMP - yval;
			XFillArc(XtDisplay(theDraw), XtWindow(theDraw), gc, 
				j*LINEWIDTH,yval,LINEWIDTH,LINEWIDTH,0,(360*64));
			XSetForeground(XtDisplay(theDraw), gc, zerox.pixel);
			XDrawLine(XtDisplay(theDraw), XtWindow(theDraw), gc,
				0, AMP, drawwidth, AMP);
	XFlush(XtDisplay(theDraw));
		}
	XFlush(XtDisplay(theDraw));
	}


	XFlush(XtDisplay(theDraw));
}
}


long randx = 1;

void srrand(x)
unsigned x;
{
	randx = x;
}


float rrand()
{
	float rval;
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	rval = (float)i/16384.0 - 1.0;
	return(rval);
}


float brrand()
{
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	return((float)i/32768.0);
}



